# Copyright 2025 West University of Timisoara
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import click

from eocube.utils.geospatialorg.cli import register as register_geospatialorg
from .auth import auth_cli

def is_installed(package_name):
    try:
        importlib.import_module(package_name)
        return True
    except ImportError:
        return False


@click.group()
def cli():
    pass


@cli.group()
def services():
    """Various Services"""
    pass


cli.add_command(services)
cli.add_command(auth_cli)

register_geospatialorg(services)

if __name__ == "__main__":
    cli()


# Copyright 2025 West University of Timisoara
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import tempfile
import uuid

import fsspec
import numpy as np
import pystac
import rasterio
import rasterio.windows
import shapely
from fsspec.implementations.local import LocalFileSystem
from shapely import box, GeometryCollection

from rocube.raster.utils import get_raster_patches


class OutputPatchHandler:
    def __init__(
        self,
        output_file_path: str,
        item: pystac.Item,
        patch_width: int = 256,
        patch_height: int = 256,
        stride: int = 256,
        assets: list = ["B02", "B03", "B04"],
        reference_asset: str = "B02",
        aoi: GeometryCollection = None,
        filesystem: fsspec.AbstractFileSystem = None,
        dtype: np.dtype = np.float32,
        shuffle: bool = False,
    ) -> None:
        """

        :param output_file_path: Destination file inside the destination filesystem (temporary file might be created)
        :param item: Input Stac Item
        :param patch_width: width of the patch
        :param patch_height: height of the patch
        :param stride: the overlap of the patches
        :param assets: list of asset names to use for data retrieval
        :param reference_asset: what asset to use as reference for generating the patches and output file
        :param aoi: Area o interest. Only patches intersecting this area are considered. Needs to be in the same projection as the data. No check is done.
        :param filesystem: Filesystem to use for uploading the result file
        :param dtype: data type of the output product
        :param shuffle: randomize the patch order
        """
        self.output_file_path = output_file_path
        self.item = item
        self.width = patch_width
        self.height = patch_height
        self.stride = stride
        self.assets = assets
        self.profile = None
        self.uuid = uuid.uuid4()
        self._tmp_file_name = f"{self.uuid.hex}.tif"
        if aoi is None:
            self.aoi = aoi
        elif isinstance(aoi, shapely.Geometry):
            self.aoi = aoi
        elif isinstance(aoi, tuple) or isinstance(aoi, list):
            self.aoi = shapely.box(*aoi)
        else:
            raise NotImplementedError(
                "AOI needs to be a shapely Geometry, a tuple or None "
            )
        self.filesystem = filesystem
        self._rios = {}
        self._index = 0  # Used by iterator
        if filesystem is None:
            self.filesystem = LocalFileSystem()
        else:
            self.filesystem = filesystem
        if reference_asset not in self.item.assets:
            raise KeyError(
                f"Reference asset: {reference_asset} not in STAC item {item.id} assets"
            )

        with rasterio.open(self.item.assets[reference_asset].href) as rio:
            self.profile = rio.profile.copy()
            self.profile["dtype"] = dtype
            self.patches = []
            for patch in get_raster_patches(
                rio, tile_width=self.width, tile_height=self.height, stride=self.stride
            ):
                bounds = rasterio.windows.bounds(
                    patch, transform=self.profile["transform"]
                )
                bbox = box(*bounds)
                if self.aoi is not None:
                    if not self.aoi.intersects(bbox):
                        continue  # Ignore box as it does not intersect with the AOI
                self.patches.append(patch)
            if shuffle:
                random.shuffle(self.patches)

        for asset in self.assets:
            if asset not in self.item.assets:
                raise KeyError(f"Asset: {asset} not in STAC item {item.id} assets")
            self._rios[asset] = rasterio.open(self.item.assets[asset].href)
        self._named_temp_file_context = tempfile.NamedTemporaryFile(
            suffix="_rocube.tif"
        )
        print(f"Destination file {self._named_temp_file_context.name}")
        self.rio = rasterio.open(
            self._named_temp_file_context.name, "w", **self.profile
        )

    def __enter__(self):
        return self

    def flush(self):
        """Function called before actual writing in order to flush any buffers

        :return:
        """
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flush()
        self.rio.close()
        with open(self._named_temp_file_context.name, "rb") as source_f:
            with self.filesystem.open(self.output_file_path, "wb") as dest_f:
                dest_f.write(source_f.read())
                print(f"Saved output to: {self.output_file_path}")

        for k, v in self._rios.items():
            if isinstance(v, rasterio.DatasetReader):
                v.close()

    def __len__(self):
        return len(self.patches)

    def __getitem__(self, idx):
        # window = self.patches[idx]
        return self._get_data(idx)

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self.patches):
            window = self.patches[self._index]
            self._index += 1
            # return self._get_data(window)
            return self._get_data(self._index)
        else:
            raise StopIteration

    def _get_data(self, idx):  # window
        window = self.patches[idx]
        bounds = rasterio.windows.bounds(window, transform=self.profile["transform"])
        rez = {"_bounds": bounds}
        for asset in self.assets:
            rio = self._rios[asset]
            win = rasterio.windows.from_bounds(*bounds, transform=rio.transform)
            data = rio.read(1, window=win)
            rez[asset] = data
        return rez

    def write(self, data: np.ndarray, idx: dict):
        """Write back the result of an processing

        :param data: Numpy array of the data to be written
        :param idx: original idx of reference patch used for generating the data. Valid only in the current context
        """
        window = self.patches[idx]
        if len(data.shape) == 2:
            self.rio.write(data, 1, window=window)
        elif len(data.shape) == 3:  # We assume channel first encoding
            channels = data.shape[0]
            for ch in range(0, channels):
                chidx = ch + 1
                self.rio.write(data[ch, ...], chidx, window=window)
        else:
            raise NotImplementedError("Unsupported shape.")


if __name__ == "__main__":
    import requests
    import tqdm

    id = "S2A_MSIL1C_20210816T092031_N0301_R093_T34TFQ_20210816T102826"

    test_url = f"https://stac.sage.uvt.ro/collections/sentinel-2-l1c/items/{id}"
    item = requests.get(test_url).json()
    stac_item = pystac.Item.from_dict(item)

    count = 0
    with OutputPatchHandler(
        f"out-{id}.tif",
        stac_item,
        patch_width=256,
        patch_height=256,
        stride=200,
        dtype=np.float32,
        shuffle=True,
    ) as result:
        for patch in tqdm.tqdm(result):
            ###
            # Do Some Processing
            data = np.zeros((256, 256), np.float32)
            data = data + count
            count += 1
            # End Processing
            ###

            # Write the result
            result.write(data, patch)



import os
import threading
import time
import logging
import click
import requests
import webbrowser
import keyring
import sys
from flask import Flask, request
from eocube.integrations.auth.client import ROCS_DISCOVERY_URL
from authlib.integrations.requests_client import OAuth2Session


EOCUBE_CLI_CLIENT_ID="eocube-cli"
EOCUBE_CLI_REDIRECT_URI="http://localhost:5123/callback"

login_done = threading.Event()

authorize_url = None
token_url = None
@click.group("auth")
def auth_cli():
    """Authentication related functionality"""


@auth_cli.command("login")
def login():
    """Login to EOCube.ro"""

    logging.getLogger('werkzeug').setLevel(logging.ERROR)

    app = Flask(__name__)
    server_thread = None


    def start_oauth_flow():
        global authorize_url
        global token_url
        oauth = OAuth2Session(EOCUBE_CLI_CLIENT_ID, redirect_uri=EOCUBE_CLI_REDIRECT_URI)
        metadata = requests.get(ROCS_DISCOVERY_URL).json()
        authorize_url = metadata["authorization_endpoint"]
        token_url = metadata["token_endpoint"]
        uri, state = oauth.create_authorization_url(
            authorize_url,
            scope='openid offline_access'
        )
        webbrowser.open(uri)
        return oauth

    def run_flask():
        app.run(port=5123, use_reloader=False)

    def shutdown_server():
        func = request.environ.get('werkzeug.server.shutdown')
        if func:
            func()

    @app.route("/callback")
    def callback():
        global oauth_token
        global token_url

        code = request.args.get("code")
        if not code:
            return "❌ No code received. Something went wrong."

        token = app.oauth_session.fetch_token(
            token_url,
            code=code
        )

        oauth_token = token  # salvează tokenul pt codul principal
        login_done.set()
        shutdown_server()  # oprește Flask


        return """
        ✅ Te-ai autentificat!<br><br>
        Poți închide fereastra.<br>
        Vezi terminalul pentru token-uri.
        """

    print("🚀 Începem loginul...")
    #global server_thread
    server_thread = threading.Thread(target=run_flask)
    server_thread.start()
    time.sleep(1)
    app.oauth_session = start_oauth_flow()
    login_done.wait()

    keyring.set_password("eocube-cli", "offline-refresh-token", oauth_token["refresh_token"])
    keyring.set_password("eocube-cli", "access-token",
                         oauth_token["access_token"])
    print("🔐 Saved tokens in keyring")
    os._exit(0)




# Copyright 2025 West University of Timisoara
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import geopandas as gpd
from geopandas import GeoDataFrame
from owslib.etree import etree
from owslib.wfs import WebFeatureService
from owslib.fes import PropertyIsEqualTo, OgcExpression, FilterRequest

COUNTY_TABLE = "geospatial:ro_judete_poligon"
UAT_TABLE = "geospatial:ro_uat_poligon"


class UAT(object):

    def __init__(
        self,
        endpoint: str = "http://www.geo-spatial.org/geoserver/ows",
        epsg: int = 4326,
    ):
        self._ows_endpoint = endpoint
        self._wfs11 = WebFeatureService(url=self._ows_endpoint, version="1.1.0")
        self._epsg = epsg

    def get_feature(
        self, feature: str, filter: OgcExpression, epsg: int = None
    ) -> GeoDataFrame:
        """Get features matching the filter. Return in specified EPSG

        :param feature: What WFS feature to query
        :param filter: The WFS filter to apply
        :param epsg: the projection the results should be
        :return: Dataframe with the result
        """
        epsg_ = epsg if epsg is not None else self._epsg
        filterxml = etree.tostring(filter.toXML(), encoding="unicode")
        response = self._wfs11.getfeature(
            feature,
            filter=filterxml,
            outputFormat="json",
            srsname=f"urn:x-ogc:def:crs:EPSG:{epsg_}",
        )
        data = gpd.GeoDataFrame.from_file(response)
        if data.empty:
            return None
        # minx, miny, maxx, maxy = data.unary_union.envelope.bounds
        return data  # , (minx, miny, maxx, maxy)


def get_administrative_unit_by_code(siruta_code: int, epsg: int = 4326) -> GeoDataFrame:
    """Query the Geo-Spatial.org service for a Romanian administrative unit by SIRUTA Code

    :param siruta_code: Siruta code of the administrative unit
    :param epsg: The EPSG of the results
    :return: GeoDataFrame with the matches
    """
    uat = UAT(epsg=epsg)
    wfs_filter = PropertyIsEqualTo(propertyname="natcode", literal=f"{siruta_code}")
    return uat.get_feature(UAT_TABLE, filter=wfs_filter)


def get_administrative_unit_by_name(
    name: str, matchcase=False, epsg: int = 4326
) -> GeoDataFrame:
    """Query the Geo-Spatial.org service for a Romanian administrative unit by name

    :param name: Name of the administrative unit
    :param matchcase: Whether to match the case
    :param epsg: The EPSG of the results
    :return: GeoDataFrame with the matches
    """
    uat = UAT(epsg=epsg)
    wfs_filter = PropertyIsEqualTo(
        propertyname="name", literal=f"{name}", matchcase=matchcase
    )
    return uat.get_feature(UAT_TABLE, filter=wfs_filter)


def get_county_by_mnemonic(county_mnemonic: str, epsg: int = 4326) -> GeoDataFrame:
    """Query the Geo-Spatial.org service for a Romanian county

    :param county_mnemonic: Mnemonic code of the county
    :param epsg: The EPSG of the results
    :return: GeoDataFrame with the matches
    """
    uat = UAT(epsg=epsg)
    wfs_filter = PropertyIsEqualTo(
        propertyname="mnemonic", literal=f"{county_mnemonic}"
    )
    return uat.get_feature(COUNTY_TABLE, filter=wfs_filter)


def get_county_by_name(name: str, epsg: int = 4326) -> GeoDataFrame:
    """Query the Geo-Spatial.org service for a Romanian county

    :param name: County Name
    :param epsg: The EPSG of the results
    :return: GeoDataFrame with the matches
    """
    uat = UAT(epsg=epsg)
    wfs_filter = PropertyIsEqualTo(propertyname="name", literal=f"{name}")
    return uat.get_feature(COUNTY_TABLE, filter=wfs_filter)


@click.command("get-county-by-name")
@click.option("--name", prompt="Name of County", help="County Name")
@click.option("--epsg", default=3844, help="EPSG of the results")
@click.option(
    "--output", default=None, help="Destination of the output saved as GeoJSON"
)
def get_county_by_name_cli(name: str, output, epsg: int = 3844):
    """Retrieves the county by name"""
    result = get_county_by_name(name, epsg=epsg)
    if result is None:
        click.echo("County not found")
        return
    if output is not None:
        result.to_file(output, driver="GeoJSON")
    else:
        click.echo(result)
    return result


@click.command("get-county-by-mnemonic")
@click.option("--name", prompt="County Mnemonic", help="County Mnemonic. Eg. TM")
@click.option("--epsg", default=3844, help="EPSG of the results")
@click.option(
    "--output", default=None, help="Destination of the output saved as GeoJSON"
)
def get_county_by_mnemonic_cli(name: str, output: str, epsg: int = 3844):
    """Retrieves the county by mnemonic"""
    result = get_county_by_mnemonic(name, epsg=epsg)
    if result is None:
        click.echo("County not found")
        return
    if output is not None:
        result.to_file(output, driver="GeoJSON")
    else:
        click.echo(result)
    return result


@click.command("get-administrative-unit-by-name")
@click.option(
    "--name", prompt="Name of administrative unit", help="Administrative unit name"
)
@click.option("--epsg", default=3844, help="EPSG of the results")
@click.option(
    "--output", default=None, help="Destination of the output saved as GeoJSON"
)
def get_administrative_unit_by_name_cli(
    name: str, output=str, matchcase=False, epsg: int = 4326
):
    """Retrieves the administrative unit by name"""
    result = get_administrative_unit_by_name(name, epsg=epsg)
    if result is None:
        click.echo("Administrative Unit not found")
        return
    if output is not None:
        result.to_file(output, driver="GeoJSON")
    else:
        click.echo(result)
    return result


@click.command("get-administrative-unit-by-code")
@click.option(
    "--code",
    prompt="Administrative unit code",
    help="SIRUTA Code of the administrative unit",
)
@click.option("--epsg", default=3844, help="EPSG of the results")
@click.option(
    "--output", default=None, help="Destination of the output saved as GeoJSON"
)
def get_administrative_unit_by_code_cli(code: int, output, epsg: int = 4326):
    """Uses the SIRUTA Code to retrieve the administrative unit"""
    result = get_administrative_unit_by_code(siruta_code=code, epsg=epsg)
    if result is None:
        click.echo("Administrative Unit not found")
        return
    if output is not None:
        result.to_file(output, driver="GeoJSON")
    else:
        click.echo(result)
    return result



# Copyright 2025 West University of Timisoara
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import requests
import threading
import xml.etree.ElementTree as ET
import datetime
from dateutil import parser as date_parser

import s3fs

from urllib.parse import urlencode
from authlib.integrations.requests_client import OAuth2Session

ROCS_DISCOVERY_URL = (
    "https://aai.eocube.ro/realms/rocs/.well-known/openid-configuration"
)
ROCS_DEFAULT_STORAGE_ENDPOINT = "https://storage.svc.uvt-01.eocube.ro"


class StorageCredentials(object):
    def __init__(
        self,
        endpoint,
        access_key,
        secret_key,
        session_token,
        expiration: datetime.datetime,
        leeway=300,
    ):
        self.endpoint = endpoint
        self.access_key = access_key
        self.secret_key = secret_key
        self.session_token = session_token
        self.expiration: datetime.datetime = expiration
        self.leeway = leeway

    def is_expired(self):
        """
        Checks whether the current token is expired based on the expiration time
        and a specified leeway.

        This method calculates the difference in seconds between the expiration time
        and the current time, adjusted for the UTC timezone. If the difference is less
        than the permissible leeway, the instance is considered expired.

        :return: Returns True if the instance is expired, otherwise False.
        :rtype: bool
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        expiration = self.expiration
        diff = expiration.timestamp() - now.timestamp()
        if diff < self.leeway:
            return True
        else:
            return False


class AuthClient(object):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        scope: list = ["openid", "email", "profile", "eocube-object-storage"],
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope
        self.metadata = requests.get(ROCS_DISCOVERY_URL).json()
        self.token_endpoint = self.metadata["token_endpoint"]
        self._token_lock = threading.Lock()
        self._storage_tokens = {}
        self._storage_tokens_lock = threading.Lock()

        self.client = OAuth2Session(
            client_id=client_id, client_secret=client_secret, scope=scope
        )

        self._token = None

    @property
    def token(self):
        if self._token is None or self._token.is_expired():
            with self._token_lock:
                self._token = self.client.fetch_token(
                    url=self.token_endpoint, grant_type="client_credentials"
                )
        return self._token

    def _credentials_are_valid(self, endpoint):
        if (
            endpoint not in self._storage_tokens
            or self._storage_tokens[endpoint].is_expired()
        ):
            return False
        else:
            return True

    def get_storage_fs(self, credentials: StorageCredentials = None):
        """Returns a fsspec filesystem for the configured credentials"""
        if credentials is None:
            credentials = self.get_storage_credentials()
        fs = s3fs.S3FileSystem(
            key=credentials.access_key,
            secret=credentials.secret_key,
            token=credentials.session_token,
            endpoint_url=credentials.endpoint,
        )
        return fs

    def get_storage_credentials(
        self, endpoint: str = None, duration_seconds: int = 3600, policy: str = None
    ):
        """
        Retrieves temporary storage credentials for a given endpoint. The credentials
        allow access to the specified storage service for a limited duration. This
        function enables fine-grained access control to the storage by optionally
        providing a policy that defines the access permissions.

        :param endpoint: The URL of the storage endpoint for which the credentials
            will be generated. If not specified, the default endpoint will be used.
        :type endpoint: str
        :param duration_seconds: The duration for which the temporary credentials
            remain valid, in seconds. If not specified, the default duration is 3600
            seconds (1 hour).
        :type duration_seconds: int
        :param policy: URL-encoded JSON-formatted policy to use as an inline session policy.
            If not provided, the policy in the JWT token will be used.
        :type policy: str
        :return: A dictionary containing the temporary credentials, including access
            keys, secret keys, and any other relevant details required to access the
            storage service.
        :rtype: dict
        """

        if endpoint is None:
            endpoint = ROCS_DEFAULT_STORAGE_ENDPOINT
        if not self._credentials_are_valid(
            endpoint
        ):  # We don't have a valid token for this endpoint
            with self._storage_tokens_lock:
                if not self._credentials_are_valid(
                    endpoint
                ):  # Double check -- due to potential concurency
                    creds = self._get_storage_credentials_using_token(
                        endpoint=endpoint,
                        token=self.token,
                        duration_seconds=duration_seconds,
                        policy=policy,
                    )
                    self._storage_tokens[endpoint] = creds
        return self._storage_tokens[endpoint]

    def _get_storage_credentials_using_access_token(
        self,
        endpoint: str,
        token: str,
        duration_seconds: int = 3600,
        policy: str = None,
    ):
        params = {
            "Action": "AssumeRoleWithWebIdentity",
            "Version": "2011-06-15",
            "WebIdentityToken": token,
            "DurationSeconds": duration_seconds,
        }

        if policy is not None:
            params["Policy"] = policy

        response = requests.post(
            endpoint,
            data=urlencode(params),
            headers={"Content-Type": "application/x-www-form-urlencoded"},
        )
        root = ET.fromstring(response.text)
        ns = {"ns": "https://sts.amazonaws.com/doc/2011-06-15/"}
        creds = root.find(".//ns:Credentials", ns)
        if creds is None:
            raise Exception("Could not parse credentials from response")

        expiration_str = creds.find("ns:Expiration", ns).text
        expiration_dt = date_parser.parse(expiration_str)

        return StorageCredentials(
            endpoint=endpoint,
            access_key=creds.find("ns:AccessKeyId", ns).text,
            secret_key=creds.find("ns:SecretAccessKey", ns).text,
            session_token=creds.find("ns:SessionToken", ns).text,
            expiration=expiration_dt,
            leeway=300,
        )

    def _get_storage_credentials_using_token(
        self,
        endpoint: str,
        token: str,
        duration_seconds: int = 3600,
        policy: str = None,
    ):
        access_token = token["access_token"]
        return self._get_storage_credentials_using_access_token(
            endpoint=endpoint,
            token=access_token,
            duration_seconds=duration_seconds,
            policy=policy,
        )


def discover_client_token():
    client_id = os.environ.get("EOCUBE_CLIENT_ID", None)
    client_secret = os.environ.get("EOCUBE_CLIENT_SECRET", None)
    client_scope = os.environ.get("EOCUBE_CLIENT_SCOPE", None)
    if client_id is None or client_secret is None:
        raise Exception("EOCUBE_CLIENT_ID and EOCUBE_CLIENT_SECRET must be set")

    kws = {"client_id": client_id, "client_secret": client_secret}
    if client_scope is not None:
        scope = re.split(r"\s+", client_scope)
        kws["scope"] = scope

    return AuthClient(**kws)
